#! /usr/bin/perl

# Create mktar.sh that creates a TAR file from the ScriptBasic sources.

# You may think that this is a simple
#
#             tar cf scriba.tar *
#
# but this is more than that. When the tarball is created the directory
# may not be fresh. There can be a lot of generated garbage that is not
# part of the source. This script reads the CV Entries files and creates
# mktar.sh that puts exactly those files into the tar archive that are source
# files.

open(BUILD,"build.txt") or die "Can not read build number";
$BUILD = <BUILD>;
close BUILD;
chomp $BUILD;

open(VERSION,"version.txt") or die "Can not read version";
$VERSION = <VERSION>;
close VERSION;
chomp $VERSION;
$DIRNAME = "scriba-${VERSION}b${BUILD}";


sub AddDirToTar {
  my $dir = shift;
  my $printdir = $dir;
  
  $printdir = "${DIRNAME}" . substr($printdir,1);

  if( ! open(IN,"$dir/CVS/Entries") ){
    warn "Can not open $dir/CVS/Entries";
    return;
    }
  my @lines = <IN>;
  close IN;
  for my $L (@lines){
    if( $L =~ m{^/(.*?)/} ){
      if( -e "$printdir/$1" ){
        print TAR "tar -rf scriba-${VERSION}b${BUILD}-source.tar $printdir/$1\n";
        }else{
        warn "$printdir/$1 is in $dir/CVS/Entries, but it does not exists.";
        }
      }elsif( $L =~ m{^D/(.*?)/} ){
      AddDirToTar("$dir/$1");
      }elsif( $L =~ m{^D\s*$} ){
      }else{
      print "Invalid line in $dir/CVS/Entries:\n>>> $L\n"
      }
    }
  }

open(TAR,">mktar") or die "Can not write mktar.sh";
print TAR <<END;
#! /bin/sh
# FILE: mktar
#
# This file was generated by scriba-tar.pl
# The program that created this file searched the
# CVS/Entries files for CV file entries and added the
# entries to this file to create the source TAR for
# ScriptBasic
#

# First of all delete the TAR file if it exists
rm scriba-${VERSION}b${BUILD}-source.tar.gz
# maybe without compression
rm scriba-${VERSION}b${BUILD}-source.tar
# create the soft link to the actual directory so that
# the tarball will contain no flat file
rm ${DIRNAME}
ln -s . ${DIRNAME}
END

&AddDirToTar('.');

print TAR <<END;
# compress the archive
gzip scriba-${VERSION}b${BUILD}-source.tar
# remove the symbolic link
rm  ${DIRNAME}
END

close TAR;
chmod(700,"mktar");

